﻿using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Random;

namespace HeuristicLab.Algorithms.IteratedSymbolicExpressionConstruction {
  [StorableClass]
  [Item("EpsGreedySymbolicExpressionConstructionPolicy", "")]
  public class EpsGreedySymbolicExpressionConstructionPolicy : SymbolicExpressionConstructionPolicyBase {

    public double Eps {
      get { return ((IFixedValueParameter<DoubleValue>)Parameters["Eps"]).Value.Value; }
      set { ((IFixedValueParameter<DoubleValue>)Parameters["Eps"]).Value.Value = value; }
    }

    public IQualityFunction QualityFunction {
      get {
        return ((IValueParameter<IQualityFunction>)Parameters["Quality function"]).Value;
      }
      set { ((IValueParameter<IQualityFunction>)Parameters["Quality function"]).Value = value; }
    }

    public EpsGreedySymbolicExpressionConstructionPolicy()
      : base() {
      Parameters.Add(new FixedValueParameter<DoubleValue>("Eps", "The fraction of random pulls", new PercentValue(0.1, true)));
      Parameters.Add(new ValueParameter<IQualityFunction>("Quality function", "The quality function to use", new TabularAvgQualityFunction()));
    }

    protected override int Select(object state, IEnumerable<int> actions, IRandom random) {
      if (random.NextDouble() < Eps) {
        return actions.SampleRandom(random, 1).First();
      }

      // find best action
      var bestActions = new List<int>();
      var bestQuality = double.NegativeInfinity;
      foreach (var a in actions) {
        double quality = QualityFunction.Q(state, a);

        if (quality >= bestQuality) {
          if (quality > bestQuality) {
            bestActions.Clear();
            bestQuality = quality;
          }
          bestActions.Add(a);
        }
      }
      return bestActions.SampleRandom(random, 1).First();
    }

    public override void Update(IEnumerable<Tuple<object, int>> stateActionSequence, double quality) {
      foreach (var t in stateActionSequence) {
        var state = t.Item1;
        var action = t.Item2;
        QualityFunction.Update(state, action, quality);
      }
    }

    protected override object CreateState(ISymbolicExpressionTreeNode root, List<int> actions, ISymbolicExpressionTreeNode parent, int childIdx) {
      return QualityFunction.StateFunction.CreateState(root, actions, parent, childIdx);
    }

    #region IItem
    protected EpsGreedySymbolicExpressionConstructionPolicy(EpsGreedySymbolicExpressionConstructionPolicy original, Cloner cloner)
      : base(original, cloner) {
    }

    [StorableConstructor]
    protected EpsGreedySymbolicExpressionConstructionPolicy(bool deserializing) : base(deserializing) { }

    public override HeuristicLab.Common.IDeepCloneable Clone(HeuristicLab.Common.Cloner cloner) {
      return new EpsGreedySymbolicExpressionConstructionPolicy(this, cloner);
    }

    #endregion
  }
}
