﻿using System;
using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Algorithms.IteratedSymbolicExpressionConstruction {
  [StorableClass]
  internal abstract class TabularQualityFunctionBase : ParameterizedNamedItem, ITabularQualityFunction {
    [Storable]
    private readonly Dictionary<object, Dictionary<int, double>> q = new Dictionary<object, Dictionary<int, double>>();
    [Storable]
    private readonly Dictionary<object, Dictionary<int, double>> qVariance = new Dictionary<object, Dictionary<int, double>>();
    [Storable]
    private readonly Dictionary<object, Dictionary<int, int>> tries = new Dictionary<object, Dictionary<int, int>>();

    public IStateFunction StateFunction {
      get {
        return ((IValueParameter<IStateFunction>)Parameters["State function"]).Value;
      }
      set { ((IValueParameter<IStateFunction>)Parameters["State function"]).Value = value; }
    }

    protected TabularQualityFunctionBase()
      : base() {
      Parameters.Add(new ValueParameter<IStateFunction>("State function", "The function that is used to map partial trees to states", new DefaultStateFunction()));
    }

    public int Tries(object state, int action) {
      Dictionary<int, int> stateTries;
      if (!tries.TryGetValue(state, out stateTries)) return 0;
      int stateActionTries;
      if (!stateTries.TryGetValue(action, out stateActionTries)) return 0;
      return stateActionTries;
    }

    public double Q(object state, int action) {
      // an action that has never been tried has q == infinity
      Dictionary<int, double> stateActionQualities;
      if (!q.TryGetValue(state, out stateActionQualities)) return double.PositiveInfinity;
      double quality;
      if (!stateActionQualities.TryGetValue(action, out quality)) return double.PositiveInfinity;
      return quality;
    }

    public double QVariance(object state, int action) {
      // an action that has never been tried has qVariance == infinity
      Dictionary<int, double> stateActionVariances;
      if (!qVariance.TryGetValue(state, out stateActionVariances)) return double.PositiveInfinity;
      double var;
      if (!stateActionVariances.TryGetValue(action, out var)) return double.PositiveInfinity;
      return var / Tries(state, action);
    }

    public virtual void Update(object state, int action, double observedQuality) {
      Dictionary<int, int> stateActionTries;
      if (!tries.TryGetValue(state, out stateActionTries)) {
        stateActionTries = new Dictionary<int, int>();
        tries.Add(state, stateActionTries);
      }
      Dictionary<int, double> stateActionQualities = null;
      if (!q.TryGetValue(state, out stateActionQualities)) {
        stateActionQualities = new Dictionary<int, double>();
        q.Add(state, stateActionQualities);
      }
      Dictionary<int, double> stateActionVariances = null;
      if (!qVariance.TryGetValue(state, out stateActionVariances)) {
        stateActionVariances = new Dictionary<int, double>();
        qVariance.Add(state, stateActionVariances);
      }
      int t;
      if (!stateActionTries.TryGetValue(action, out t)) {
        t = 0;
        stateActionTries.Add(action, t + 1);
        stateActionQualities.Add(action, observedQuality);
        stateActionVariances.Add(action, 0); // naive initialization
      } else {
        stateActionTries[action] = t + 1;
        var delta = observedQuality - stateActionQualities[action];
        var curMean = CalculateNewQ(state, action, observedQuality);
        stateActionQualities[action] = curMean;
        stateActionVariances[action] = stateActionVariances[action] + delta * (observedQuality - curMean); // iterative calculation of mean
      }
    }

    protected abstract double CalculateNewQ(object state, int action, double observedQuality);


    #region item
    [StorableConstructor]
    protected TabularQualityFunctionBase(bool deserializing) : base(deserializing) { }
    protected TabularQualityFunctionBase(TabularQualityFunctionBase original, Cloner cloner)
      : base(original, cloner) {
      this.q = new Dictionary<object, Dictionary<int, double>>(original.q);
      this.qVariance = new Dictionary<object, Dictionary<int, double>>(original.qVariance);
      this.tries = new Dictionary<object, Dictionary<int, int>>(original.tries);
    }
    #endregion

    public void InitializeState() {
      ClearState();
    }

    public void ClearState() {
      q.Clear();
      tries.Clear();
      qVariance.Clear();
    }
  }
}
